
/******************************************************************************
*
*               J_Sort V1.0 2013/12/03
*
*   File name:          j_script_sort.c
*   Version:            V1.0
*   Date:               3rd December 2013
*   ISDC calls:         RILlogMessage
*						DAL3GENelapsedOBT
*   Developer:          Sami Maisala & Eero Esko
*						Carol Anne Oxborrow
*						Lucia Pavan, ISDC ( lucia.pavan@unige.ch )
*   Contact Address:    Observatory
*                       P.O Box 14
*                       Tahtitorninmaki
*                       00014 University of Helsinki
*                       Finland
*			Also: oxborrow@space.dtu.dk
*   Purpose:            To sort an instrument gain history table: Reads 
*                       timeList and orderList arrays and sorts orderList
*                       with quicksort algorithm by OBTimes in timeList
*
*   Revision History
*
*   Date        Version Programmer      What changed
*
*   2013/12/03	1.0	LP	SCREW-01150: First version of this function, 
*				based on j_cor_gain_sort_history.C v.8.1
*
*   Return Values:
*   Value                       Meaning
*   ISDC_OK             Function completed without problem
*   Negative            Function terminated without completing due to an error
*                       
*
*
* 
*   Parameter arguments:     
*
*   Name         Data Type       I/O     Description          Valid Range
*
*   timeList    *OBTime        Input/   Pointer to OBTime    DAL defined OBTime
*				Output  Output array is ordered
*   orderList   *int		Input/  Pointer to orderList  0...numPoints
*				Output  Output array contains original indices
*					of the ordered times ( 0 to numEvts-1)
*   nEvtRowStart int		Input	First row in list to begin sort
*					To sort entire list nEvtRowStart=0
*   nEvtRowEnd	 int		Input	Last row in list to end sort
*					To sort whole list nEvtRowEnd=numEvts-1
*   status       int            Input   Usual input status   0 or negative
*                                                      
*
*******************************************************************************
*/     

// #include <j_cor_gain.h>
// #include <jemx.h>

int j_sort(OBTime* timeList,
			    int *orderList,
			    int  nEvtRowStart,
			    int  nEvtRowEnd,
			    int status)
{
  int  nHiLoPartition=0;

  if (nEvtRowStart < nEvtRowEnd) {
    nHiLoPartition = j_partition(timeList, orderList,
				     nEvtRowStart, nEvtRowEnd, status);
    if ( nHiLoPartition<0 ){
    	RILlogMessage(NULL, Warning_2,
    	    "%d: Couldn't sort OBTs in gain history table", nHiLoPartition);
    	return status;
    }
				     
    j_sort(timeList, orderList, nEvtRowStart, nHiLoPartition,
			    status);
			    
    j_sort(timeList, orderList, nHiLoPartition + 1, nEvtRowEnd,
			    status);
  }
  return status;
}


int j_partition( OBTime *timeList,
				 int *orderList,
				 int  nStart,
				 int  nEnd,
				 int status)
{
  long loCountUp = nStart - 1;
  long hiCountDown = nEnd + 1;
  long nRandom=0;

  OBTime pivot=0;
  OBTime timeswap=0;
  OBTime elapsedOBT=0;

  int modeswap;

  /* Randomizing the "pivot" element to make the runtime of quicksort \
     independent of the input array */
  
  
  nRandom = DAL3GENrandomInt() % (nEnd - nStart + 1) + nStart;
  pivot = timeList[nRandom];
  status=ISDC_OK;

  do {
    do {
      hiCountDown--;
      status = DAL3GENelapsedOBT( timeList[hiCountDown], 
	    		pivot, &elapsedOBT, status);
      if ( status!=ISDC_OK ){
      	  RILlogMessage(NULL, Warning_2,
      	  	"%d: Couldn't sort OBT times of gain history table", status);
      	  return status;
      }
    } while ( elapsedOBT>0 );
    do {
      loCountUp++;
      status = DAL3GENelapsedOBT( pivot, 
	    		timeList[loCountUp], &elapsedOBT, status);
      if ( status!=ISDC_OK ){
      	  RILlogMessage(NULL, Warning_2,
      	  	"%d: Couldn't sort OBT times of gain history table", status);
      	  return status;
      }
    } while ( elapsedOBT>0 );
    if (loCountUp < hiCountDown) {
      
      timeswap = timeList[loCountUp];
      timeList[loCountUp] = timeList[hiCountDown];
      timeList[hiCountDown] = timeswap;
      
      modeswap = orderList[loCountUp];
      orderList[loCountUp] = orderList[hiCountDown];
      orderList[hiCountDown] = modeswap;


    } else {
      return hiCountDown;
    }
  } while (1);
  
  return status;
  
}


