//////////////////////////////////////////////////////////////////////////////
//
//  File name:  jemx_scw_analysis.C
//  Version:    see below
//  Component:  jemx_scw_analysis
//  Developer:  Peter Kretschmar, ISDC
//              Peter.Kretschmar@obs.unige.ch
//              Silvia Martnez Nez, GACE, Valencia University
//              Silvia.Martinez@uv.es
//              Simona Soldi, ISDC,
//              Simona.Soldi@unige.ch
//              Lucia Pavan, ISDC
//              Lucia.Pavan@unige.ch
//  Purpose:
//
//  Execute a complete scientific analysis of a single JEM-X Science Window,
//  beginning with data corrections and finishing with images, spectra and
//  lightcurves - or a user-selected subset of these steps.
//
//  History:
//  
//  1.0 - 2001-06-15 - First version for ISSW Consolidation
//  1.1 - 2001-07-11 - First formally delivered version
//  2.0 - 2001-09-12 - Updated script with isdcroot 3.1 improvements
//  2.1 - 2001-11-12 - Common high-level script parameters
//  2.2 - 2002-05-10 - Included new parameters, adapted to new libscript
//  3.0 - 2002-09-25 - Function jemx_check_sw now also checks if a minimum
//                     set of data on a given ScW exists [SMN+PK]
//                     Modified handling of analysis levels [PK]
//  3.1 - 2002-11-07 - Include level LCR (j_src_extract_lc)
//                     Some parameters moved to general
//  3.2 - 2002-11-08 - Allow analysis of non-pointing ScWs [PK]
//  3.3 - 2002-11-21 - Allow GUI for jemx_scw_analysis [SMN]
//  3.4 - 2002-12-12 - Implemented advanced GUI Capabilities [SMN]
//  3.5 - 2002-12-27 - Allow for direct instMod parameter [PK]
//  3.6 - 2003-05-03 - Include call to idx_find on gain history [PK]
//  3.7.0 -2003-06-19 - Give all scripts same version 
//  3.7.3 - 2003-09-17 - Set radiusLimit to BIN_S and BIN_T steps [SMN]
//  3.8.0 - 2003-10-09 - SCREW 1209 - Aligned start times [SMN]
//  3.8.1 - 2003-10-15 - SPR 3249 - Error status of ScW analysis not cleared [PK]
//  3.9.0 - 2004-01-27 - instMod parameter for j_bin_spectra
//  4.0.0 - 2004-03-22 - Check for FULL-* instead of FULL-RAW [PK]
//  4.1.0 - 2004-04-21 - Use independent j_dead instead of j_dead_time
//  4.2.0 - 2004-05-12 - Add parameter outputExists for COR and DEAD (SPR 3563)
//  4.2.1 - 2004-06-04 - Added missing parameter IMA_fineResol (SPR 3666) [PK]
//  4.3.0 - 2004-09-03 - Allow for different subdir structure (SCREW 1532) [PK]
//  4.3.2 - 2004-11-02 - Allow for negative LCR_timeStep (SPR 3893) [PK]
//  4.4.1 - 2004-12-01 - Correct parameter numPixelSubDiv (SPR 3939) [PK]
//  4.5.6 - 2005-09-16 - startLevel default value set to COR (par file only) (SPR 4192) [SMN] 
//  4.6   - 2006-09-04 - GUI gives unexpected error message when quit 
//                       button is pressed (SPR 04364) [SMN]
//        - 2006-09-08 - BKG level removed from j_scripts (SCREW 01896) [SMN]
//        - 2006-09-19 - swg keyword ISDCLEVL sets correctly (SPR-04413) [SMN]
//  6.0.0 - 2008-11-17 - Convert SPE_timeStep parameter to BIN_S_timeStep as part
// 			 of the changes to replace j_src_spectra with j_reform_spectra 
//                       (SCREW-2094)[So]
//			 Remove BKG step from GUI (SPR 4775)[So]
//  6.1   - 2009-04-13 - SCREW 2134 [SP]
//  6.2.3 - 2011-10-18 - SCREW 2326 [LP] Adding the nChanBins < 0 standard definition
//						(versioning in parallel with j_scripts version)
//  6.3   -2012-05-11  - SPR 5041 [LP] pass parameters timeStart and timeStop to j_src_extract_spectra
//  6.4   -2013-07-30  - SCREW 2502 [LP] exclude from spectra analysis firsts SCWs after reactivation (perigee).
//  6.4.1 -2014-03-21  - SPR 05073 [LP] initialization of arrays "low_index" and "high_index" in function check_skipSCW
//  6.4.2 - 2014-12-08 - SPR 05086 [NM] fixed a bug for mac mavericks on j_rebin_rmf
//  6.4.3 - 2015-02-19 - SCREW 2572 [LP] include an option to analyse also SLEW science windows (ScwType)
//  6.4.4 - 2016-01-18 - SPR 05104 [LP] make the scripts case insensitive to the param ScwType (strncmp --> strncasecmp in jemx_check_sw)
//                       SPR 05105 [LP] updated the check_skipSCW funxtion to recognise cases when the HK data do not include the final perigee period
//  6.5.0  -2016-02-24 [LP] update of param files to work with new j_ima_iros (new LC feature included)
//
//////////////////////////////////////////////////////////////////////////////

// Ensure standard setup
#define ISDCMAIN jemx_scw_analysis
#define MAIN_VER "6.5.0"
#include "scriptstuff.h"
#include <sys/types.h>
#include <dirent.h>

// Instrument-specific error codes defined by JEM-X SDAST
#define J_SCW_SCRIPT_ERR_START         (J_ERROR_CODE_START  - 82000)
#define J_SCW_BAD_DETECTOR_NUMBER      (J_SCW_SCRIPT_ERR_START -  1)
#define J_SCW_BAD_LEVEL_DEFINITION     (J_SCW_SCRIPT_ERR_START -  2)
// Specific code to report to calling script if data is insufficient
#define J_SCW_NO_MINIMUM_DATA          (J_SCW_SCRIPT_ERR_START - 99)

//Internal specific error codes
#define J_SCW_NOT_POINTING             (J_SCW_SCRIPT_ERR_START - 98)

// verbosity levels
#define J_CHATTY_VERY_QUIET   0 /* Only system messages and fatal errors */
#define J_CHATTY_QUIET        1 /* Errors and severe warnings only */
#define J_CHATTY_NORMAL       2 /* + important Log messages, extend warnings */
#define J_CHATTY_VERBOSE      3 /* + program flow comments */
#define J_CHATTY_VERY_VERBOSE 4 /* + miscellaneous program control values */
#define J_CHATTY_MAXIMAL      5 /* Everything you want */

// Default energy bands for nChanBins < 1
static const char *StdChanMin0="46";
static const char *StdChanMax0="223";

static const char *StdChanMin1="46 130";
static const char *StdChanMax1="129 223";

static const char *StdChanMin2="46 77 130 175";
static const char *StdChanMax2="76 129 174 223";

static const char *StdChanMin3="46 59 77 102 130 153 175 199";
static const char *StdChanMax3="58 76 101 129 152 174 198 223";

static const char *StdChanMin4="46 52 59 68 77 89 102 117 130 141 153 165 175 187 199 211";
static const char *StdChanMax4="51 58 67 76 88 101 116 129 140 152 164 174 186 198 210 223";

static const char *StdChanMin5="46 49 52 55 59 63 68 72 77 83 89 95 102 109 117 123 130 135 141 147 153 159 165 170 175 181 187 193 199 205 211 217";
static const char *StdChanMax5="48 51 54 58 62 67 71 76 82 88 94 101 108 116 122 129 134 140 146 152 158 164 169 174 180 186 192 198 204 210 216 223";

static const char *StdChanMin6="46 49 51 53 55 57 59 61 63 65 67 69 71 73 75 77 79 81 84 87 90 93 96 99 103  106 110 114 118 121 124 127 130 132 135 139 142 145 148 151 154 157 160 162 166 168 171 173 176 179 182 185 188 191 194 197 200 203 206 209 212 215 218 221";
static const char *StdChanMax6="48 50 52 54 56 58 60 62 64 66 68 70 72 74 76 78 80 83 86 89 92 95 98 102 105 109 113 117 120 123 126 129 131 134 138 141 144 147 150 153 156 159 161 165 167 170 172 175 178 181 184 187 190 193 196 199 202 205 208 211 214 217 220 223";

static const char *StdChanMin7="0 2 4 6 8 10 12 14 16 18 20 22 24 26 28 30 32 34 36 38 40 42 44 46 48 50 52 54 56 58 60 62 64 66 68 70 72 74 76 78 80 82 84 86 88 90 92 94 96 98 100 102 104 106 108 110 112 114 116 118 120 122 124 126 128 130 132 134 136 138 140 142 144 146 148 150 152 154 156 158 160 162 164 166 168 170 172 174 176 178 180 182 184 186 188 190 192 194 196 198 200 202 204 206 208 210 212 214 216 218 220 222 224 226 228 230 232 234 236 238 240 242 244 246 248 250 252 254";
static const char *StdChanMax7="1 3 5 7 9 11 13 15 17 19 21 23 25 27 29 31 33 35 37 39 41 43 45 47 49 51 53 55 57 59 61 63 65 67 69 71 73 75 77 79 81 83 85 87 89 91 93 95 97 99 101 103 105 107 109 111 113 115 117 119 121 123 125 127 129 131 133 135 137 139 141 143 145 147 149 151 153 155 157 159 161 163 165 167 169 171 173 175 177 179 181 183 185 187 189 191 193 195 197 199 201 203 205 207 209 211 213 215 217 219 221 223 225 227 229 231 233 235 237 239 241 243 245 247 249 251 253 255";



// Prototype for utility function;
using namespace std;
using namespace ISDC;
int jemx_reportTaskStatus(ISDCTask *task);

int jemx_check_sw(const char *swgDOL,
		  int        jemxNum,
		  int        chatter,
		  double     *startTimeSpe,
		  double     *startTimeLcr,
		  double     SpetimeStep,
		  double     LcrtimeStep,
		  char       *subDirCor,
		  char       *subDirRes,
                  const char *ScwType,
		  int        status);

int jemx_used_levels(const char *startLevel,
		     const char *endLevel,
		     const char *skipLevels,
		     char *usedLevels,
		     int  chatter,
		     int  callStatus);

int findSubDir(const char *dir, const char *subdir);

int check_skipSCW( const char *swgDOL,
			int jemxNum,
			int chatter,
			bool& skipSCW,
			char swID[20],
			int callStatus);

int j_sort(OBTime *timeList,
		   int *orderList,
		   int  nEvtRowStart,
		   int  nEvtRowEnd,
		   int status);

int j_partition(OBTime *timeList,
				int *orderList,
				int  nStart,
				int  nEnd,
				int status);

//*************************************************************************
// Now to the actual script which includes the following steps:
// - Set up script as ISDCTask 
// - Set up called lower level scripts as ISDCTasks 
// - Determine start and end level of analysis
// - Execute the individual analysis levels if they fall within the range
//   defined by the user. The levels and associated lower level scripts are:
//
//   COR   - j_correction            Correct data for instr. fingerprints
//   GTI   - j_gti                   Derive Good-Time Intervals
//   DEAD  - j_dead                  Calculate deadtime factors 
//   CAT_I - j_cat_extract           Extract list of expected sources
//   BIN_I - j_image_bin             Bin events to shadowgrams
//   IMA   - j_imaging               Deconvolve images, find sources
//   SPE   - j_src_extract_spectra   Extract spectra of individual sources
//   LCR   - j_src_extract_lc        Extract lc of individual sources
//   BIN_S - j_bin_spectra           Bin overall detector spectra
//   BIN_T - j_bin_lc                Bin overall detector lightcurves
//
//*************************************************************************

int jemx_scw_analysis(int argc=0, char **argv=0)
{
    // Use standard namespaces
    using namespace std;
    using namespace ISDC;

    // The following line is there to have the GUI pop up

    // Local variable definitions
    int        status=ISDC_OK;             // processing status 
    ISDCLevel  startLevel=NO_LEVEL;        // level at which analysis begins  
    ISDCLevel  endLevel=NO_LEVEL;          // level at which analysis ends
    char       rilString[DAL_FILE_NAME_STRING]; // utility string for RIL calls
    char       usedLevels[DAL_MAX_STRING]; // used for level handling
    char       subDirCor[DAL_MAX_STRING];  // holds relative sub directory
    char       subDirRes[DAL_MAX_STRING];  // holds relative sub directory
    Instrument jemxInstr=JMX1;             // detector code (JMX1 or JMX2)
    double     startTimeSpe=0.0;           // global time Start (SPE/BIN_S)
    double     startTimeLcr=0.0;           // global time Start (LCR/BIN_T)
    char       chanLowStr[DAL_FILE_NAME_STRING];
    char       chanHighStr[DAL_FILE_NAME_STRING];
    int        nChanBins=0;
	
	double		 timeDel;	//time difference between SCW start time and REV perigee time
	bool		 skipSCW ;  // boolean flag. Y= the current SCW is in a period close to reactivation. Gain is still variable.
	char		 swID[20]={0};// current SCWID number
	
    // one-pass do loop for error control
    do {

    //
	// ==== Set up script ===============================================
	//
	// The script itself is set up as ISDCTask object. This ensures
	// correct setup of some ISDC library functions and simplifies
        // access to the script parameters within the script itself.
	//

	ISDCTask jemx_scw_analysis(ISDCMAINNAME,
				   MAIN_VER,
				   argc,argv,
				   ISDCTask::kPARAMGUI |ISDCTask::kCLI);
	status = jemx_scw_analysis.GetStatus();

	if (status==kUSER_ABORT){
	    sprintf(rilString,"Quitting upon User Request.");
	    RILlogMessage(NULL,Warning_0,rilString);
	    break;
	}

	if (ISDC_OK != status) { 
	    sprintf(rilString,"Could not set up script! status = %d",status);
	    RILlogString(NULL,Error_2,rilString);
	    break; 
	}


	// ==== Check the given Science Window ==============================
	//
	// The given Science Window Group is checked for the following 
	// properties:
	// - it is a pointing;
	// - it is for JMX1 or JMX2
	// - it contains JMX[12] raw data
	// Also a relative timeStart can be set
	// and it is determined where corrected data & results go
	//

	startTimeSpe=(double)jemx_scw_analysis["timeStart"];
	startTimeLcr=(double)jemx_scw_analysis["timeStart"];

	status=jemx_check_sw((const char*)jemx_scw_analysis["swgDOL"], 
			     (int) jemx_scw_analysis["general_jemxNum"],
			     (int) jemx_scw_analysis["general_chatter"],
			     &startTimeSpe,
			     &startTimeLcr,
			     (double) jemx_scw_analysis["BIN_S_timeStep"],
			     (double) jemx_scw_analysis["LCR_timeStep"],
			     subDirCor,
			     subDirRes,
                             (const char*) jemx_scw_analysis["ScwType"],
			     status);

        if (ISDC_OK!=status) { break; } 


	// ==== Set up level handling ======================================
	startLevel = 
	    IRlevelGetCode((const char*)jemx_scw_analysis["startLevel"]);
	// Check if startLevel is reasonable
	if (NO_LEVEL==startLevel) {
	    sprintf(rilString,"Level name '%s' not recognized!",
		    (const char*)jemx_scw_analysis["startLevel"]);
	    RILlogString(NULL,Error_2,rilString);
	    return(J_SCW_BAD_LEVEL_DEFINITION);
	}

	endLevel = 
	    IRlevelGetCode((const char*)jemx_scw_analysis["endLevel"]);
	//Check if endLevel is reasonable
	if (NO_LEVEL==endLevel) {
	    sprintf(rilString,"Level name '%s' not recognized!",
		    (const char*)jemx_scw_analysis["endLevel"]);
	    RILlogString(NULL,Error_2,rilString);
	    return(J_SCW_BAD_LEVEL_DEFINITION);
	}
        
	status = jemx_used_levels((const char*)jemx_scw_analysis["startLevel"],
				  (const char*)jemx_scw_analysis["endLevel"],
				  (const char*)jemx_scw_analysis["skipLevels"],
				  usedLevels,
				  (int)jemx_scw_analysis["general_chatter"],
				  status);
        if (ISDC_OK!=status) { break; } 

	//
	// ==== Set up lower level scripts ==================================
	//
	// The lower level scripts are set up as ISDCTask objects.
	// If anything goes wrong with this, we break to the end.
	
	ISDCTask j_gti("j_gti");
	status = jemx_reportTaskStatus(&j_gti);
	if (ISDC_OK != status) { break; }

	ISDCTask j_dead("j_dead");
	status = jemx_reportTaskStatus(&j_dead);
	if (ISDC_OK != status) { break; }

	ISDCTask j_cat_extract("j_cat_extract");
	status = jemx_reportTaskStatus(&j_cat_extract);
	if (ISDC_OK != status) { break; }

	ISDCTask j_image_bin("j_image_bin");
	status = jemx_reportTaskStatus(&j_image_bin);
	if (ISDC_OK != status) { break; }

	ISDCTask j_imaging("j_imaging");
	status = jemx_reportTaskStatus(&j_imaging);
	if (ISDC_OK != status) { break; }

	ISDCTask j_src_extract_spectra("j_src_extract_spectra");
	status = jemx_reportTaskStatus(&j_src_extract_spectra);
	if (ISDC_OK != status) { break; }

	ISDCTask j_src_extract_lc("j_src_extract_lc");
	status = jemx_reportTaskStatus(&j_src_extract_lc);
	if (ISDC_OK != status) { break; }

	ISDCTask j_bin_spectra("j_bin_spectra");
	status = jemx_reportTaskStatus(&j_bin_spectra);
	if (ISDC_OK != status) { break; }

	ISDCTask j_bin_lc("j_bin_lc");
	status = jemx_reportTaskStatus(&j_bin_lc);
	if (ISDC_OK != status) { break; }

	//
	// ==== Set up parameters of ScW analysis script ====================
	//
		
	nChanBins=static_cast<int>(jemx_scw_analysis["nChanBins"]);
	strcpy(chanLowStr,(const char *)jemx_scw_analysis["chanLow"]);
	strcpy(chanHighStr,(const char *)jemx_scw_analysis["chanHigh"]);
		
	  //
          // --- Define standard energy bins (passed with parameter nChanBins < 0)   -------
	  //
	  // Note: These values are moved here from jemx_science_analysis.C, 
	  //       to be able to run also jemx_scw_analysis with standard energy 
	  //       bins without launching jemx_science_analysis (used e.g. by the JEMX team)
	  //       2011-10-18 - SCREW 2326
	  //
 switch (nChanBins)
	{
		case 0:
			nChanBins = 1;
			strcpy(chanLowStr,StdChanMin0);
			strcpy(chanHighStr,StdChanMax0);
			break;
		case -1:
			nChanBins = 2;
			strcpy(chanLowStr,StdChanMin1);
			strcpy(chanHighStr,StdChanMax1);
			break;
		case -2:
			nChanBins = 4;
			strcpy(chanLowStr,StdChanMin2);
			strcpy(chanHighStr,StdChanMax2);
			break;
		case -3:
			nChanBins = 8;
			strcpy(chanLowStr,StdChanMin3);
			strcpy(chanHighStr,StdChanMax3);
			break;
		case -4:
			nChanBins = 16;
			strcpy(chanLowStr,StdChanMin4);
			strcpy(chanHighStr,StdChanMax4);
			break;
		case -5:
			nChanBins = 32;
			strcpy(chanLowStr,StdChanMin5);
			strcpy(chanHighStr,StdChanMax5);
			break;
		case -6:
			nChanBins = 64;
			strcpy(chanLowStr,StdChanMin6);
			strcpy(chanHighStr,StdChanMax6);
			break;
		case -7:
			nChanBins = 128;
			strcpy(chanLowStr,StdChanMin7);
			strcpy(chanHighStr,StdChanMax7);
			break;
	    }

	    jemx_scw_analysis["nChanBins"] = nChanBins;
	    jemx_scw_analysis["chanLow"]   = chanLowStr;
	    jemx_scw_analysis["chanHigh"]  = chanHighStr;

        
        
	//
	// ==== Run analysis ===============================================
	// Call the lower level scripts one after the other if their
	// associated analysis level is within the selection of used levels
	//

	// CORrection step
	if(strstr(usedLevels,"COR")) {

	    // Do not correct OSim data
	    if ((int)jemx_scw_analysis["general_osimData"]) {
		RILlogString(NULL,Log_1,
			     "j_correction not executed for OSim data.");
	    } else {

		ISDCTask j_correction("j_correction");
		status = jemx_reportTaskStatus(&j_correction);
		if (ISDC_OK != status) { break; }

		j_correction.CopyPars(jemx_scw_analysis,"general_");
		j_correction.CopyPars(jemx_scw_analysis,"COR_");
		j_correction.CopyPars(jemx_scw_analysis,"IC_",ISDCTask::kKEEP);
		j_correction["swgDOL"]  = jemx_scw_analysis["swgDOL"];
		j_correction["instMod"] = jemx_scw_analysis["instMod"];
		j_correction["subDir"]  = subDirCor;
		status = j_correction.Run();
	    
		if (ISDC_OK == status) {
		    RILlogString(NULL,Log_1,
				 "Successfully finished level COR "
				 "- data corrections");
		} else {
		    RILlogString(NULL,Error_2,
				 "Error in level COR - data corrections!");
		    break;    // on error go to end
		}
	    }
	}

	// Good Time Intervals
	if(strstr(usedLevels,"GTI")) {

	    j_gti.CopyPars(jemx_scw_analysis,"general_");
	    j_gti.CopyPars(jemx_scw_analysis,"GTI_");
	    j_gti.CopyPars(jemx_scw_analysis,"IC_",ISDCTask::kKEEP);
	    j_gti["swgDOL"] = jemx_scw_analysis["swgDOL"];
		j_gti["subDir"] = subDirRes;
	    status = j_gti.Run();

	    if (ISDC_OK == status) {
		RILlogString(NULL,Log_1,
			      "Successfully finished level GTI - good time intervals");
	    } else {
		RILlogString(NULL,Error_2,
			      "Error in level GTI - good time intervals");
		break;    // on error go to end
	    }
	}

	// DEADtime history
	if(strstr(usedLevels,"DEAD")) {

	    j_dead.CopyPars(jemx_scw_analysis,"general_");
	    j_dead.CopyPars(jemx_scw_analysis,"DEAD_");
	    j_dead.CopyPars(jemx_scw_analysis,"IC_",ISDCTask::kKEEP);
	    j_dead["swgDOL"]  = jemx_scw_analysis["swgDOL"];
	    j_dead["instMod"] = jemx_scw_analysis["instMod"];
	    j_dead["subDir"]  = subDirRes;
	    status = j_dead.Run();

	    if (ISDC_OK == status) {
		RILlogString(NULL,Log_1,
			      "Successfully finished level DEAD - deadtime calculation");
	    } else {
		RILlogString(NULL,Error_2,
			      "Error in level DEAD - deadtime calculation!");
		break;    // on error go to end
	    }
	}
	
	// CATalog extraction for Imaging
	if(strstr(usedLevels,"CAT_I")) {

	    j_cat_extract.CopyPars(jemx_scw_analysis,"general_");
	    j_cat_extract.CopyPars(jemx_scw_analysis,"CAT_I_");

	    j_cat_extract.CopyPars(jemx_scw_analysis,"IC_",ISDCTask::kKEEP);
	    j_cat_extract["swgDOL"] = jemx_scw_analysis["swgDOL"];
	    j_cat_extract["subDir"] = subDirRes;
	    status = j_cat_extract.Run();

	    if (ISDC_OK == status) {
		RILlogString(NULL,Log_1,
			      "Successfully finished level CAT_I - catalog extraction");
	    } else {
		RILlogString(NULL,Error_2,
			      "Error in level CAT_I - catalog extraction");
		break;    // on error go to end
	    }
	}

	// BINning for Imaging
	if(strstr(usedLevels,"BIN_I")) {

	    j_image_bin.CopyPars(jemx_scw_analysis,"general_");
	    j_image_bin.CopyPars(jemx_scw_analysis,"BIN_I_");
	    j_image_bin.CopyPars(jemx_scw_analysis,"IC_",ISDCTask::kKEEP);
	    j_image_bin["swgDOL"]      = jemx_scw_analysis["swgDOL"];
	    j_image_bin["subDir"]      = subDirRes;
	    j_image_bin["nChanBins"]   = jemx_scw_analysis["nChanBins"];
	    j_image_bin["chanLow"]     = jemx_scw_analysis["chanLow"];
	    j_image_bin["chanHigh"]    = jemx_scw_analysis["chanHigh"];
	    j_image_bin["timeStart"]   = jemx_scw_analysis["timeStart"];
	    j_image_bin["timeStop"]    = jemx_scw_analysis["timeStop"];
	    j_image_bin["radiusLimit"] = jemx_scw_analysis["radiusLimit"];
	    j_image_bin["instMod"]     = jemx_scw_analysis["instMod"];
	    j_image_bin["gtiNames"]    = jemx_scw_analysis["GTI_MergedName"];

	    status = j_image_bin.Run();

	    if (ISDC_OK == status) {
		RILlogString(NULL,Log_1,
			     "Successfully finished level BIN_I - "
			     "shadowgram binning");
	    } else {
		RILlogString(NULL,Error_2,
			     "Error in level BIN_I - shadowgram binning!");
		break;    // on error go to end
	    }
	}

	//
	// verify if the current SCW is coming shortly after a reactivation period.
	// In this case the gain is highly variable, and the calibration not optimal.
	// Issue a warning even at IMA level, and avoid SPE level 
	//  
				
		status=check_skipSCW( (const char*)jemx_scw_analysis["swgDOL"], 
						(int) jemx_scw_analysis["general_jemxNum"],
						(int) jemx_scw_analysis["general_chatter"],
						skipSCW,   //output
						swID,
						status);
		// if something went wrong issue a warning but go on with the analysis
		if (ISDC_OK!=status) { 
			skipSCW = false ;
			status = ISDC_OK ;
			RILlogString(NULL,Warning_1,
					"An error occured when checking for instrument switch-off periods");
			sprintf(rilString,
					"The instrument will be considered on. status %d",status);
			RILlogString(NULL,Warning_1,rilString);						 
		}
		
		
	// IMAges and source lists
		
	if(strstr(usedLevels,"IMA")) {
		// if the current SCW is coming shortly after a reactivation period, 
		// do the IMA step but issue a warning in the log file
		
	    if(skipSCW && !jemx_scw_analysis["skipIMAfirstScw"]){
			RILlogString(NULL,Warning_1,"**************************************************************");
			sprintf(rilString,"SCW %s is following a reactivation period", swID);
			RILlogString(NULL,Warning_1,rilString);
			RILlogString(NULL,Warning_1,"Be aware of possible issues with GAIN calibration for this SCW.");
			RILlogString(NULL,Warning_1,"**************************************************************");
		}
		
		if(!skipSCW || !jemx_scw_analysis["skipIMAfirstScw"]){
			j_imaging.CopyPars(jemx_scw_analysis,"general_");
			j_imaging.CopyPars(jemx_scw_analysis,"IMA_");
			j_imaging.CopyPars(jemx_scw_analysis,"IC_",ISDCTask::kKEEP);
			j_imaging["swgDOL"]      = jemx_scw_analysis["swgDOL"];
			j_imaging["subDir"]      = subDirRes;
			j_imaging["instMod"]     =  jemx_scw_analysis["instMod"];
			//	    j_imaging["radiusLimit"] = jemx_scw_analysis["radiusLimit"];

/// all LC-related params are kept in the front end at LC step/tab.
/// if the user asks to have LC with new j_ima_iros, then pass the relevant LCR params to j_imaging and j_ima_iros
			if (jemx_scw_analysis["LCR_useIROS"]){
				///LCR_timeStep,r,h,4,0.1,, "[seconds] should be > 0.1, defaults to 4 s"
				if ( double(jemx_scw_analysis["LCR_timeStep"]) >= 0.1) {
					j_imaging["timeStep"]      = jemx_scw_analysis["LCR_timeStep"];
				}
				j_imaging["makeLCs"] = "yes";
				j_imaging["doBurstSearch"]      = jemx_scw_analysis["LCR_doBurstSearch"];
				j_imaging["referenceTime"]      = jemx_scw_analysis["LCR_referenceTime"];
				j_imaging["burstThreshSrc"]      = jemx_scw_analysis["LCR_burstThreshSrc"];
				j_imaging["burstThreshDet"]      = jemx_scw_analysis["LCR_burstThreshDet"];
			} else {
				j_imaging["makeLCs"] = "no";
			}


			status = j_imaging.Run();
			
			if (ISDC_OK == status) {
				RILlogString(NULL,Log_1,
							 "Successfully finished level IMA - image deconvolution");
			} else {
				RILlogString(NULL,Error_2,
							 "Error in level IMA - image deconvolution!");
				break;    // on error go to end
			}
			
		}else{
			RILlogString(NULL,Warning_1,"**************************************************************");
			sprintf(rilString,"SCW %s is following a reactivation period. ",swID);
			RILlogString(NULL,Warning_1,rilString);
			RILlogString(NULL,Warning_1,"You prevented IMA extraction with parameter skipIMAfirstScw = Yes");
			RILlogString(NULL,Warning_1,"No IMA extraction will be performed");
			RILlogString(NULL,Warning_1,"**************************************************************");
		}
	}

		
	// SPEctrum extraction for individual sources

	if(strstr(usedLevels,"SPE")) {
		// if the current SCW is coming shortly after a reactivation period,
		// avoid SPE extraction (unless the user has explicetely forced the 
		// extraction with skipSPEfirstScw=NO		
		
		if(skipSCW && !jemx_scw_analysis["skipSPEfirstScw"]){
			RILlogString(NULL,Warning_1,"**************************************************************");
			sprintf(rilString,"SCW %s is following a reactivation period.", swID);
			RILlogString(NULL,Warning_1,rilString);
			RILlogString(NULL,Warning_1,"You forced SPE extraction with parameter skipSPEfirstScw = No");
			RILlogString(NULL,Warning_1,"Be aware of possible issues with GAIN calibration for this SCW.");
			RILlogString(NULL,Warning_1,"**************************************************************");			
		}
		
		
		if(!skipSCW || !jemx_scw_analysis["skipSPEfirstScw"]){
			j_src_extract_spectra.CopyPars(jemx_scw_analysis,"general_");
			j_src_extract_spectra.CopyPars(jemx_scw_analysis,"SPE_");
			j_src_extract_spectra.CopyPars(jemx_scw_analysis,"IC_",ISDCTask::kKEEP);
			j_src_extract_spectra["swgDOL"]    = jemx_scw_analysis["swgDOL"];
			j_src_extract_spectra["subDir"]    = subDirRes;
			j_src_extract_spectra["response"]  =  jemx_scw_analysis["response"];
			j_src_extract_spectra["arf"]       =  jemx_scw_analysis["arf"];
			//SPR 5041 - pass parameters timeStart(startTimeSpe) and timeStop to j_src_extract_spectra
			j_src_extract_spectra["timeStart"] =  startTimeSpe;
			j_src_extract_spectra["timeStop"]  =  jemx_scw_analysis["timeStop"];
			
			status = j_src_extract_spectra.Run();
			
			if (ISDC_OK == status) {
				RILlogString(NULL,Log_1,
							 "Successfully finished level SPE - source spectra");
			} else {
				RILlogString(NULL,Error_2,
							 "Error in level SPE - source spectra!");
				break;    // on error go to end
			}
		}else{

			RILlogString(NULL,Warning_1,"**************************************************************");
			sprintf(rilString,"SCW %s is following a reactivation period. ",swID);
			RILlogString(NULL,Warning_1,rilString);
			RILlogString(NULL,Warning_1,"The gain is too variable for a proper calibration ");
			RILlogString(NULL,Warning_1,"              No SPECTRAL extraction will be performed");
			RILlogString(NULL,Warning_1,"If you want to force spectral extraction use parameter skipSPEfirstScw = 'N' ");
			RILlogString(NULL,Warning_1,"but be aware of possible issues with GAIN calibration for this SCW.");
			RILlogString(NULL,Warning_1,"**************************************************************");
		}
	}
 		
	// LightCuRve extraction for individual sources
	if(strstr(usedLevels,"LCR")) {

		// if the current SCW is coming shortly after a reactivation period, 
		// do the LCR step but issue a warning in the log file
		if(skipSCW && !jemx_scw_analysis["skipLCRfirstScw"]){
			RILlogString(NULL,Warning_1,"**************************************************************");
			sprintf(rilString,"SCW %s is following a reactivation period",swID);
			RILlogString(NULL,Warning_1,rilString);
			RILlogString(NULL,Warning_1,"Be aware of possible issues with GAIN calibration for this SCW.");
			RILlogString(NULL,Warning_1,"**************************************************************");			
		}
	    
		if(!skipSCW || !jemx_scw_analysis["skipLCRfirstScw"]){

		    if (!jemx_scw_analysis["LCR_useIROS"]){

			j_src_extract_lc.CopyPars(jemx_scw_analysis,"general_");
			//j_src_extract_lc.CopyPars(jemx_scw_analysis,"LCR_");
			j_src_extract_lc.CopyPars(jemx_scw_analysis,"IC_",ISDCTask::kKEEP);
			j_src_extract_lc["swgDOL"]    = jemx_scw_analysis["swgDOL"];
			j_src_extract_lc["subDir"]    = subDirRes;
			j_src_extract_lc["nChanBins"] = jemx_scw_analysis["nChanBins"];
			j_src_extract_lc["chanLow"]   = jemx_scw_analysis["chanLow"];
			j_src_extract_lc["chanHigh"]  = jemx_scw_analysis["chanHigh"];
			j_src_extract_lc["timeStart"] = startTimeLcr;
			j_src_extract_lc["timeStop"]  = jemx_scw_analysis["timeStop"];
			j_src_extract_lc["timeStep"]  = jemx_scw_analysis["LCR_timeStep"];
			j_src_extract_lc["radiusLimit"] = jemx_scw_analysis["radiusLimit"];
			j_src_extract_lc["nPhaseBins"]= jemx_scw_analysis["nPhaseBins"];
			j_src_extract_lc["phaseBins"] = jemx_scw_analysis["phaseBins"];
			j_src_extract_lc["gtiNames"]  = jemx_scw_analysis["GTI_MergedName"];
			j_src_extract_lc["instMod"]   =  jemx_scw_analysis["instMod"];
			j_src_extract_lc["response"]   =  jemx_scw_analysis["response"];

			/// SCREW ... feb 2016 [LP]
			/// part of the LCR_* parameters defined at jemx_scw_analysis are instead related to the new j_ima_iros LC extraction tool.
			/// therefore don't use any more teh automatic LCR_** copying, but copy only the params realted to the old LC software
			j_src_extract_lc["vignCorr"]	=jemx_scw_analysis["LCR_vignCorr"];
			j_src_extract_lc["evtType"]		=jemx_scw_analysis["LCR_evtType"];
			j_src_extract_lc["precisionLevel"]	=jemx_scw_analysis["LCR_precisionLevel"];
			j_src_extract_lc["tAccuracy"]	=jemx_scw_analysis["LCR_tAccuracy"];
			j_src_extract_lc["useRaDec"]	=jemx_scw_analysis["LCR_useRaDec"];
			j_src_extract_lc["rowSelect"]	=jemx_scw_analysis["LCR_rowSelect"];
			j_src_extract_lc["fluxScaling"]	=jemx_scw_analysis["LCR_fluxScaling"];
			j_src_extract_lc["skipNearDeadAnode"]=jemx_scw_analysis["LCR_skipNearDeadAnode"];
			j_src_extract_lc["skipHotSpot"]	=jemx_scw_analysis["LCR_skipHotSpot"];
			j_src_extract_lc["overrideCollTilt"]=jemx_scw_analysis["LCR_overrideCollTilt"];


			status = j_src_extract_lc.Run();
			
			if (ISDC_OK == status) {
				RILlogString(NULL,Log_1,
			      "Successfully finished level LCR - source lightcurves");
			} else {
				RILlogString(NULL,Error_2,
			      "Error in level LCR - source lightcurves!");
				break;    // on error go to end
			}
		   }
		} else{
			RILlogString(NULL,Warning_1,"**************************************************************");
			sprintf(rilString,"SCW %s is following a reactivation period. ",swID);
			RILlogString(NULL,Warning_1,rilString);
			RILlogString(NULL,Warning_1,"You prevented LCR extraction with parameter skipLCRfirstScw = Yes");
			RILlogString(NULL,Warning_1,"No LCR extraction will be performed");
			RILlogString(NULL,Warning_1,"**************************************************************");
		}
	}

	// BINning data to detector Spectra
	if(strstr(usedLevels,"BIN_S")) {

	    j_bin_spectra.CopyPars(jemx_scw_analysis,"general_");
	    j_bin_spectra.CopyPars(jemx_scw_analysis,"BIN_S_");
	    j_bin_spectra.CopyPars(jemx_scw_analysis,"IC_",ISDCTask::kKEEP);
	    j_bin_spectra["gtiNames"]    = jemx_scw_analysis["GTI_MergedName"];
	    j_bin_spectra["swgDOL"]      = jemx_scw_analysis["swgDOL"];
	    j_bin_spectra["subDir"]      = subDirRes;
	    j_bin_spectra["timeStart"]   = startTimeSpe;
	    j_bin_spectra["timeStop"]    = jemx_scw_analysis["timeStop"];
	    j_bin_spectra["nPhaseBins"]  = jemx_scw_analysis["nPhaseBins"];
	    j_bin_spectra["phaseBins"]   = jemx_scw_analysis["phaseBins"];
	    j_bin_spectra["response"]    = jemx_scw_analysis["response"];
	    j_bin_spectra["instMod"]     = jemx_scw_analysis["instMod"];
            j_bin_spectra["radiusLimit"] = jemx_scw_analysis["radiusLimit"];
	    status = j_bin_spectra.Run();

	    if (ISDC_OK == status) {
		RILlogString(NULL,Log_1,
			     "Successfully finished level BIN_S - detector spectra");
	    } else {
		RILlogString(NULL,Error_2,
			      "Error in level BIN_S - detector spectra!");
		break;    // on error go to end
	    }
	}

	// BINning data to detector lightcurves for Timing analysis
	if(strstr(usedLevels,"BIN_T")) {

	    j_bin_lc.CopyPars(jemx_scw_analysis,"general_");
	    j_bin_lc.CopyPars(jemx_scw_analysis,"BIN_T_");
	    j_bin_lc.CopyPars(jemx_scw_analysis,"IC_",ISDCTask::kKEEP);
	    j_bin_lc["swgDOL"]    = jemx_scw_analysis["swgDOL"];
	    j_bin_lc["subDir"]    = subDirRes;
	    j_bin_lc["gtiNames"]  = jemx_scw_analysis["GTI_MergedName"];
	    j_bin_lc["nChanBins"] = jemx_scw_analysis["nChanBins"];
	    j_bin_lc["chanLow"]   = jemx_scw_analysis["chanLow"];
	    j_bin_lc["chanHigh"]  = jemx_scw_analysis["chanHigh"];
	    j_bin_lc["timeStart"]  = startTimeLcr;
	    j_bin_lc["timeStop"]   = jemx_scw_analysis["timeStop"];
	    j_bin_lc["timeStep"]   = jemx_scw_analysis["LCR_timeStep"];
	    j_bin_lc["nPhaseBins"] = jemx_scw_analysis["nPhaseBins"];
	    j_bin_lc["phaseBins"]  = jemx_scw_analysis["phaseBins"];
	    j_bin_lc["instMod"]    =  jemx_scw_analysis["instMod"];
		j_bin_lc["radiusLimit"] = jemx_scw_analysis["radiusLimit"];
	    status = j_bin_lc.Run();

	    if (ISDC_OK == status) {
		RILlogString(NULL,Log_1,
			     "Successfully finished level BIN_T "
			     "- detector lightcurves");
	    } else {
		RILlogString(NULL,Error_2,
			      "Error in level BIN_T - detector lightcurves!");
		break;    // on error go to end
	    }
	}


	// Mark SWG with last accepted level
         IRgroupNameSetLevel((const char *)jemx_scw_analysis["swgDOL"],
                             endLevel);

    } while (0); 
    // end of single-pass do loop


    // If the only reason for breaking was a ScW which was not a Pointing,
    // then reset the status. Report case of no data to calling script.

    if (J_SCW_NOT_POINTING==status) {
	status = ISDC_OK; 
    }

    return(status);
}

//******************************************************************
//
// Function jemx_reportTaskStatus
//
//******************************************************************

int jemx_reportTaskStatus(ISDCTask *task) 
{
    using namespace std;
    using namespace ISDC;
    int   status=ISDC_OK;
    char  rilString[DAL_MAX_STRING];

    status = task->GetStatus();
    if (ISDC_OK!=status) {
	sprintf(rilString,"Problem with task - its status is %d!",
		status);
//		task.TaskName(),status);
	RILlogString(NULL,Error_2,rilString);
    } else {
	sprintf(rilString,"Task %s set up without problems",
		task->TaskName());
	RILlogString(NULL,Log_0,rilString);
    }

    return(status);
}





//////////////////////////////////////////////////////////////////////////////
// Function jemx_check_sw
//
// Does sanity checks on given Science Window:
// - Is it of the same type as specfied in ScwType
// - Is it a JEM-X ScW of the correct detector number
//
//////////////////////////////////////////////////////////////////////////////


int jemx_check_sw(const char *swgDOL,
				  int        jemxNum,
				  int        chatter, 
				  double     *startTimeSpe,
				  double     *startTimeLcr,
				  double     spetimeStep,
				  double     lcrtimeStep,
				  char       *subDirCor,
				  char       *subDirRes,
                  		  const char *ScwType,
				  int        status)

{
	
	char         rilString[DAL_FILE_NAME_STRING];
	int          foundRaw=0;                   // checks existence of raw data
	char         dsName[14];                   // used for JMXi-XXXX-RAW|ALL
	char         relPath[DAL_FILE_NAME_STRING]; // relative path to/from SWG
	char         subPath[DAL_FILE_NAME_STRING]; // relative path from SWG
	
	sw_type      swType=OTHER;                 // from ScW types in DAL3GEN  
	Instrument   jemxInst=INST_NOT_RECOGNIZED; // from instruments in DAL3GEN
	dal_element *swgPtr=NULL;                  // handle to the SW Group
	dal_element *childPtr=NULL;                // handle to a SW Group child
	
	double      swStartT=0.0;  //TSTART SW
	double      swStopT=0.0;   //TSTOP  SW
	double      Nspe=0.0;         //Multiplicative factor SPE-BIN_S
	double      Nlcr=0.0;         //Multiplicative factor LCR-BIN_T
	
	
	if (ISDC_OK!=status) { return(status); }
	
	do { //Error handling loop
		
		status=DALobjectOpen (swgDOL,&swgPtr,status);
		
		if (ISDC_OK != status) { 
			sprintf(rilString,
					"Could not open SW %s %d",swgDOL, status);
			RILlogString(NULL,Error_2,rilString);
			break; 	                         
		}
		
		// Somewhat of an overkill, but we use the generic DAL3GEN
		// function to be sure of doing this right. Also this allows
		// us to easily add other checks in the future.
		
		status=DAL3GENgetSWG(swgPtr,0,NULL,NULL,&swType,NULL,&jemxInst,
							 NULL,NULL,NULL,NULL,status);
		if (ISDC_OK != status) { 
			sprintf(rilString,
					"Error accessing data of SW %s %d",swgDOL, status);
			RILlogString(NULL,Error_2,rilString);
			break; 	                         
		}

        


        if (J_CHATTY_VERBOSE<=chatter) {
        	sprintf(rilString,"input ScwType: %s",ScwType);
       		RILlogString(NULL,Log_0,rilString);
        	sprintf(rilString,"current SCW Type: %d  (POINTING=0, SLEW=1, OTHER=2)",
                	swType);
        	RILlogString(NULL,Log_0,rilString);
        }
        
        
        if ( (strncasecmp( ScwType, "ALL" ,3) == 0) ||
	     (strncasecmp( ScwType, "POI" ,3) == 0 && (POINTING==swType)) ||
	     (strncasecmp( ScwType, "SLEW" ,3) == 0 && (SLEW==swType))
            ){
		//if (POINTING==swType)
			if ( (1==jemxNum && JMX1==jemxInst) ||  // Check instrument
				(2==jemxNum && JMX2==jemxInst) ) {
				if (J_CHATTY_VERBOSE<=chatter) {    // Report if OK
					sprintf(rilString,
							"The given ScW is a Pointing with JMX%1d data.",
							jemxNum);
					RILlogString(NULL,Log_0,rilString);
				}
			} else {
				status = J_SCW_BAD_DETECTOR_NUMBER;
				sprintf(rilString,
						"The given ScW does not contain data for JMX%1d! %d",
						jemxNum,status);
				RILlogString(NULL,Error_2,rilString);
				break;
			}
		} else {     // Only ScWs of the type required by the user are accepted by the S/W!
			status=J_SCW_NOT_POINTING;

            sprintf(rilString,"The given ScW is not of type %s - "
			"no analysis will be performed on it.", ScwType);
	    RILlogString(NULL,Log_2,rilString);
	    break;
	   }
		
		// OK, now lets see if we can find any of the raw data types
		// JMXi-FULL-RAW, -REST-RAW, -SPTI-RAW, -TIME-RAW or -SPEC-RAW
		// We assume that if these exist, the connected data is there as well.
		foundRaw=0;
		
		sprintf(dsName,"JMX%1d",jemxNum);
		
		strcpy(dsName+4,"-FULL-*");
		status=DALobjectFindElement(swgPtr,dsName,&childPtr,status);
		if (ISDC_OK==status) { 
			foundRaw++;
		} else if (DAL_ELEMENT_NOT_FOUND==status){ // No problem, just continue
			status = ISDC_OK;
		} else if(ISDC_OK!=status){
			sprintf(rilString,"jemx_check_sw(): Error searching %s in SWG! " 
					"Status= %d",dsName,status);
			RILlogString(NULL,Error_2,rilString);
			break;
		}
		
		strcpy(dsName+4,"-REST-*");
		status=DALobjectFindElement(swgPtr,dsName,&childPtr,status);
		if (ISDC_OK==status) { 
			foundRaw++;
		} else if (DAL_ELEMENT_NOT_FOUND==status){ // No problem, just continue
			status = ISDC_OK;
		} else if(ISDC_OK!=status){
			sprintf(rilString,"jemx_check_sw(): Error searching %s in SWG! " 
					"Status= %d",dsName,status);
			RILlogString(NULL,Error_2,rilString);
			break;
		}
		
		strcpy(dsName+4,"-SPTI-*");
		status=DALobjectFindElement(swgPtr,dsName,&childPtr,status);
		if (ISDC_OK==status) { 
			foundRaw++;
		} else if (DAL_ELEMENT_NOT_FOUND==status){ // No problem, just continue
			status = ISDC_OK;
		} else if(ISDC_OK!=status){
			sprintf(rilString,"jemx_check_sw(): Error searching %s in SWG! " 
					"Status= %d",dsName,status);
			RILlogString(NULL,Error_2,rilString);
			break;
		}
		
		strcpy(dsName+4,"-TIME-*");
		status=DALobjectFindElement(swgPtr,dsName,&childPtr,status);
		if (ISDC_OK==status) { 
			foundRaw++;
		} else if (DAL_ELEMENT_NOT_FOUND==status){ // No problem, just continue
			status = ISDC_OK;
		} else if(ISDC_OK!=status){
			sprintf(rilString,"jemx_check_sw(): Error searching %s in SWG! " 
					"Status= %d",dsName,status);
			RILlogString(NULL,Error_2,rilString);
			break;
		}
		
		strcpy(dsName+4,"-SPEC-*");
		status=DALobjectFindElement(swgPtr,dsName,&childPtr,status);
		if (ISDC_OK==status) { 
			foundRaw++;
		} else if (DAL_ELEMENT_NOT_FOUND==status){ // No problem, just continue
			status = ISDC_OK;
		} else if(ISDC_OK!=status){
			sprintf(rilString,"jemx_check_sw(): Error searching %s in SWG! " 
					"Status= %d",dsName,status);
			RILlogString(NULL,Error_2,rilString);
			break;
		}
		
		if (1>foundRaw) {
			sprintf(rilString,
					"Available data in the ScW insufficient to perform "
					"Scientific Analysis");
			RILlogString(NULL,Warning_2,rilString);
			status=J_SCW_NO_MINIMUM_DATA;
			break;
		} else if (J_CHATTY_VERBOSE<=chatter) {    // Report if OK
			sprintf(rilString,"Found raw data for JMX%1d",jemxNum);
			RILlogString(NULL,Log_0,rilString);
		}
		
		//Getting the ScW TSTART & TSTOP
		
		status = DALattributeGet(swgPtr,"TSTART",DAL_DOUBLE,
								 (void*)&swStartT,NULL,NULL,status);
		if (ISDC_OK != status) {
			sprintf(rilString,
					"jemx_check_scw(): Could not get start time of ScW! status=%d",
					status);
			RILlogMessage(NULL,Error_2,rilString);
			return(status);
		}
		
		status = DALattributeGet(swgPtr,"TSTOP",DAL_DOUBLE,
								 (void*)&swStopT,NULL,NULL,status);
		if (ISDC_OK != status) {
			sprintf(rilString,
					"jemx_check_scw(): Could not get stop time of ScW! status=%d",
					status);
			RILlogMessage(NULL,Error_2,rilString);
			return(status);
		}
		
		
		//If TSTOP is less than TSTART - NO DATA.
		
		if(swStartT >= swStopT){
			
			sprintf(rilString,
					"No data in the ScW. TSTART greater or equal TSTOP");
			RILlogString(NULL,Warning_2,rilString);
			status=J_SCW_NO_MINIMUM_DATA;
			break;
			
		}
		
		//If timeStart is greater than TSTOP - NO DATA.
		
		if(*startTimeSpe >= swStopT){
			
			sprintf(rilString,
					"No data in the requested time interval for the ScW."
					"Requested startTime is greater or equal than TSTOP");
			RILlogString(NULL,Warning_2,rilString);
			status=J_SCW_NO_MINIMUM_DATA;
			break;
			
		}
		
		//If startTime is less than TSTART - We need to redefine startTime
		
		
		if(*startTimeSpe < swStartT && spetimeStep > 0 && *startTimeSpe > 0){
			
			modf((swStartT-(*startTimeSpe))/spetimeStep,&Nspe);
			*startTimeSpe=Nspe*spetimeStep+(*startTimeSpe);
			
			if (J_CHATTY_VERBOSE<=chatter) {    
				sprintf(rilString,"Redefined BIN_S startTime %lf"
						"User has given a start Time is less than SW TSTART",
						*startTimeSpe);
				RILlogString(NULL,Log_0,rilString);
			}
			
		}else{
			
			*startTimeSpe=*startTimeSpe;
			
			if (J_CHATTY_VERBOSE<=chatter) {    
				sprintf(rilString,"BIN_S startTime is: %lf (-1 means" 
						"SW bounds are taking", *startTimeSpe);
				RILlogString(NULL,Log_0,rilString);
			}
			
		}
		
		
		if(*startTimeLcr < swStartT && lcrtimeStep > 0 && *startTimeLcr > 0){
			
			modf((swStartT-(*startTimeLcr))/lcrtimeStep,&Nlcr);
			*startTimeLcr=Nlcr*lcrtimeStep+(*startTimeLcr);
			
			if (J_CHATTY_VERBOSE<=chatter) {    
				sprintf(rilString,"Redefined LCR and BIN_T startTime %lf" 
						"User has given a start Time is less than SW TSTART", 
						*startTimeSpe);
				RILlogString(NULL,Log_0,rilString);
			}   
			
			
		}else {
			
			*startTimeLcr=*startTimeLcr;
			
			if (J_CHATTY_VERBOSE<=chatter) {    
				sprintf(rilString,"LCR and BIN_T startTime is: %lf (-1 means" 
						"SW bounds are taking", *startTimeSpe);
				RILlogString(NULL,Log_0,rilString);
			}
			
		}
		
		//
		// Check for the existence of cor/ and res/ subdirectories
		// and set subDirXXX to these, otherwise empty.
		// This should handle both rev_1 and rev_2 directory structures.
		//
		
		status = DALparseDOL(swgDOL,NULL,NULL,NULL,NULL,relPath,
							 NULL,NULL,NULL,NULL,NULL,status);
		if (ISDC_OK != status) {
			sprintf(rilString,
					"Could not parse SWG DOL %s! status=%d",
					swgDOL,status);
			RILlogMessage(NULL,Error_2,rilString);
			break;
		}
		
		// Check out subDir jmx[12]
		sprintf(subPath,"%s/jmx%1d",relPath,jemxNum);
		
		if (findSubDir(subPath,"cor")) { 
			sprintf(subDirCor,"jmx%1d/cor",jemxNum);
		} else {
			strcpy(subDirCor,"");
		}
		if (J_CHATTY_VERBOSE<=chatter) {    
			sprintf(rilString,"Writing corrected data to %s%s",
					relPath,subDirCor);
			RILlogString(NULL,Log_0,rilString);
		}
		
		if (findSubDir(subPath,"res")) { 
			sprintf(subDirRes,"jmx%1d/res",jemxNum);
		} else {
			strcpy(subDirRes,"");
		}
		if (J_CHATTY_VERBOSE<=chatter) {    
			sprintf(rilString,"Writing analysis results to %s%s",
					relPath,subDirRes);
			RILlogString(NULL,Log_0,rilString);
		}
		
		// end of error-control loop
	} while(0);
	
	// Close SWG if no major error has occured
	if (ISDC_OK==status ||
		J_SCW_NOT_POINTING==status ||
		J_SCW_NO_MINIMUM_DATA==status) {
		DALobjectClose(swgPtr,DAL_SAVE,status);
	}
	return status;
	
}

//***************************************************************************
// 
//   Function jemx_set_used_levels
//
//   Description:
//      
//     Returns a string with the names of the levels actually used,
//     based on the start level, the end level and the information
//     about skipped levels.
//     This string is used by the main script in deciding if a given
//     level will be executed.
//
//  Error codes:
//
//     0 == ISDC_OK   ==> function executed with no errors
//     < 0 (various)  ==> ISDC system function error code, passed on
//
//  Parameter arguments:
//     name           range    I/O     description
//
//     startLevel     string   input   first level to execute
//     endLevel       string   input   last level to execute
//     skipLevels     string   input   levels to be skipped  
//     usedLevels     string   output  levels finally used
//     chatter        0-5      input   verbosity parameter
//     callStatus     any      input   status on input
//
//***************************************************************************
int jemx_used_levels(const char *startLevel,
					 const char *endLevel,
					 const char *skipLevels,
					 char *usedLevels,
					 int  chatter,
					 int  callStatus)
{
    char allLevels[]="COR,GTI,DEAD,CAT_I,BIN_I,IMA,SPE,LCR,BIN_S,BIN_T";
    char skipString[DAL_MAX_STRING];
    char rilString[DAL_FILE_NAME_STRING]; 
    char *skipLevel=NULL;
    char *subStart=NULL;
    char *subEnd=NULL;
    int  status=ISDC_OK;
    
    // Exit immediately if calling status is not OK
    if (ISDC_OK != callStatus) {return(callStatus);} 
	
    // 1. Take level range between startLevel and endLevel ----------------
    subStart=strstr(allLevels,startLevel);
    subEnd=strstr(allLevels,endLevel);
    
    if (NULL==subStart || NULL==subEnd) {
		return(J_SCW_BAD_LEVEL_DEFINITION);
    }
	
    //    We want the range up to the next comma or end of string
    subEnd=strstr(subEnd,",");
    if (NULL==subEnd) {
		subEnd = allLevels + strlen(allLevels) + 1;
    }
    //    Copy subset of allLevels to usedLevels
    strncpy(usedLevels,subStart,subEnd-subStart);
    usedLevels[subEnd-subStart] = '\0';
	
    // 2. Remove all skipped levels ---------------------------------------
    if (0==strlen(skipLevels)) { // leave now if we have nothing to skip
		return(ISDC_OK);
    }
    
    strcpy(skipString,skipLevels);
    //    Use strtok to parse work string into individual level chunks,
    //    e.g. "COR, IMA" -> "COR" "IMA"
	
    skipLevel=strtok(skipString,",; ");
	
    // get position of level to skip and position just after it
    subStart=strstr(usedLevels,skipLevel);
    if (NULL==subStart) {                     // not found in usedLevels
		subStart=strstr(allLevels,skipLevel); // check if it is known at all
		if (NULL==subStart) {  
			sprintf(rilString,
					"Skipped analysis %s is unknown!",skipLevel);
			RILlogString(NULL,Warning_1,rilString);
		} else if (J_CHATTY_VERBOSE <= chatter) { // if yes we are not in range
			sprintf(rilString,
					"Skipped level %s is not in range of %s - %s !",
					skipLevel,startLevel,endLevel);
			RILlogString(NULL,Warning_1,rilString);
		}
    } else {
		subEnd=strstr(subStart,",");
		// Now remove the selected piece. If we have a following comma
		// we copy everything after that, otherwise we just send an
		// end mark _before_ subStart
		if (NULL==subEnd) {
			*(subStart-1) = '\0';
		} else {
			strcpy(subStart,subEnd+1);	    
		}
    }
	
    while ( (skipLevel=strtok(NULL,",; ")) != NULL ) {
		
		// get position of level to skip and position just after it
		subStart=strstr(usedLevels,skipLevel);
		if (NULL==subStart) {
			sprintf(rilString,
					"Analysis level name %s to be skipped not included!",
					skipLevel);
			RILlogString(NULL,Warning_2,rilString);
		} else {
			subEnd=strstr(subStart,",");
			
			// Now remove the selected piece. If we have a following comma
			// we copy everything after that, otherwise we just send an
			// end mark _before_ subStart
			if (NULL==subEnd) {
				*(subStart-1) = '\0';
			} else {
				strcpy(subStart,subEnd+1);	    
			}
		}
    }
	
    if (J_CHATTY_VERBOSE<=chatter) {
		sprintf(rilString,"Analysis levels to be executed: %s",
				usedLevels);
		RILlogString(NULL,Log_1,rilString);
    }
	
	
    return(status);
}	

int findSubDir(const char *dir, const char *subdir){
  DIR *d=opendir(dir);
  if (d!=NULL){
    dirent* f;
    while(f=readdir(d)){
	if (!strcmp(f->d_name,subdir)) {
	    return(1);
	}
    }
    closedir(d);
  }
  return(0);
}


//***************************************************************************
//
// Function check_skipSCW
// Author     : Lucia Pavan
//              lucia.pavan@unige.ch
//
// Description:
//		verify the current SCW start time, 
//		and compare this to the REV perigee time.
//		This is used to filter out from SPE analysis the firsts SCWs 
//		after reactivation (perigee)
//		For these SCWs the gain is highly variable and the 
//		calibration can not be optimal.
//
//
// Input: 
//		swgDOL, jemxNum, chatter
//
// Output: 
//		skipSCW is a boolean. Generally it is "NO". It is set to "YES" only for SCWs too close to a reactivation period
//		swID (used in jemx_scw_analysis for the warning message in the log)
//      status
//
//***************************************************************************

int check_skipSCW( const char *swgDOL,
				   int jemxNum,
				   int chatter,
				   bool& skipSCW, //output
				   char swID[20], //output
				   int status)   //output
{
	char rilString[DAL_FILE_NAME_STRING];
	char osmCnv[300];		// pointer to the OSM converted file
	char relPathOsm [DAL_FILE_NAME_STRING];		//path to the hk OSM files
	char filePathOsm [DAL_FILE_NAME_STRING];	//file name of the hk OSM files
	char extNameOsm [DAL_FILE_NAME_STRING];		//extension of the hk OSM files
	
	int	revNUM=0;			// current REV number
	int	revolNum = 0;		//used for check while reading the OSM data
	int	varLen=0;
	int	*histOrderPtr = NULL;   
    int *hvPruneList = NULL;   // pointer cotaining the oerdered list for HV__DELTA values
    int numDisorder = 0;
	
    long i=0;
    long j=0;
    long k=0;
    long iSort=0;
    long iTest=0;
    long nZero=0;
    long nNan=0;
	long nStart=0;
    long nStop=0;
    long numRead = 0;
    long bufferSize = 0;
	long numOsmPoints = 0;
	long numUsePoints = 0;
	long binSize = 0;
	long numIntervals=NULL; //number of intervals when the instruments were switched off.
	long low_index[10]={0} ;	//array storing the lower boundaries for the periods
							// when the instruments were switched off
	long high_index[10]={0} ;	//array storing the upper boundaries for the periods
							// when the instruments were switched off
	for (i = 0; i < 10; ++i)
	{
	  low_index[i] = -1; //initialize indexes to -1
	  high_index[i] = -1;
	}

	dal_dataType dataType = DAL_DOUBLE;
	dal_dataType diskType;			// the data type will be defined as output by DALtableGetColStruct
	dal_element  *swgPtr=NULL;      // pointer to the SW Group
	dal_element  *dalPtrOsm=NULL;	// DAL pointer used for OSM data

	OBTime  scwOBTstart;			// start time of the current SCW (in OBT format)
	OBTime  scwOBTend;				// end time of the current SCW (in OBT format)
    OBTime	elapsedOBT = 0;			//dummy variable used to compute time intervals
    OBTime  binOBT = 0;
	OBTime  deltaOBT = 0 ;			//elapsed OBT time for each interval of instrument switch-off
	OBTime *orderedOBTs = NULL;
	OBTime *osmOBTbuffer = NULL;	//buffer containing all the OBT times read in from the OSM data.
	float  *hvDeltaBuffer = NULL;	//buffer containing the HV__DELTA values read in from the OSM data.
    double *HVdata = NULL; 
	
	//const values to flag and order data 
	int	const good_osm = 1;		// Pruning flag indicating data and OBTs read in are all good
	int const obt_nan = -1;		// Pruning flag for OBTs that are NAN
	int const obt_zero = -2;	// Pruning flag for OBTs that are zero or missing
	int const osm_nan = -3;		// Pruning flag for any OSM value that's a NAN
	int const osm_zero = -4;	// Pruning flag for any OSM value that's zero, invalid or missing
	double const osm_hvLOWER = 0.00;		// Lowest valid HV__DELTA value
	double const osm_hvUPPER = 1.00e+03;	// Highest valid HV__DELTA value
	double const HVswitch=200.0 ;	//threshold HV value to consider the instruments on/off
	
 #define LIN_CONST deltaOBT  /* how much time after instrument switch-on the gain 
							is considered stable again (so that SPE analysis can be done)
							LIN_CONST = deltaOBT means that it takes the same interval as 
							the instrument switch-off period to recover the gain stability */
	

	
 do { //one-pass do loop for error control
	
	//open the SWGgroup
	status=DALobjectOpen(swgDOL,&swgPtr,status);

	if (ISDC_OK != status) { 
		sprintf(rilString,
			"Could not open SW %s %d", swgDOL, status);
		RILlogString(NULL,Error_1,rilString);
		break;
	}

	//read for this SWGgroup the SCWid, REVnum, start/end time of the SCW
	status=DAL3GENgetSWG(swgPtr,0,NULL,swID,NULL,NULL,NULL,
					 &revNUM,NULL,&scwOBTstart,&scwOBTend,status);
	if (ISDC_OK != status) { 
		sprintf(rilString,
			"Error finding SWID %s %d",swID, status);
		RILlogString(NULL,Error_1,rilString);
		break;
	} else { 
		if (J_CHATTY_VERBOSE<= chatter) {    // Report if OK
			sprintf(rilString,
				"swID: %s, revNUM: %d, OBT start: %lld, end: %lld, status: %d", 
				swID, revNUM, scwOBTstart,scwOBTend, status);
			RILlogString(NULL,Log_1,rilString);
		}
	}
	
	// Close SWG
	if (ISDC_OK==status) {
		DALobjectClose(swgPtr,DAL_SAVE,status);
	}
	
	
	//
	//	    Find the OSM data
	//
	
	sprintf(relPathOsm,"/scw/%04d/rev.VVV/osm",revNUM);
	sprintf(filePathOsm,"jmx%1d_avg_hk",jemxNum);
	sprintf(extNameOsm,"JMX%1d-CSSW-AVG",jemxNum);
	
	status = DAL3GENopenRepFile(relPathOsm,filePathOsm,extNameOsm,&dalPtrOsm,status);
	if (ISDC_OK!=status) { 
		sprintf(rilString,
				"Could not find ../../%s/%s(%s)! status = %d",
				relPathOsm,filePathOsm,extNameOsm,status);
		RILlogString(NULL,Error_1,rilString);
		break;
	} 
	else {
		status = DALelementGetLocation(dalPtrOsm,osmCnv,status);
		sprintf(rilString,
				"Using OSM converted data DOL: %s", osmCnv);
		RILlogString(NULL,Log_0,rilString);
	}
	 
	//
	//	    Check we're looking at OSM data for the correct revolution
	//
	status = DALattributeGet(dalPtrOsm, "REVOL", DAL_INT, 
							 (void *)&revolNum, NULL, NULL, status);
	
	 if ( status!=ISDC_OK ){
		RILlogMessage(NULL, Warning_2,
					  "Couldn't get REVOL keyword from OSM AVG extension: %d",
					  status);
	}
	 if ( revolNum!= revNUM){
		RILlogMessage(NULL, Error_1,
					  "OSM extension belongs to wrong revolution: %d, should be %d",
					  revolNum, revNUM );
		break;
	}
	
	//	    Find out about buffer requirements for the OBT column
	//
	status = DALtableGetColStruct(dalPtrOsm, 0, &numOsmPoints,
								  (char*) "OB_TIME", &dataType, &binSize, &varLen, status);

	 if (status != ISDC_OK){
		RILlogMessage(NULL, Error_1,
					  "GetColStruct failed for OSM data OB_TIME: %d", status);
		break;
	}

	if (J_CHATTY_VERBOSE<= chatter) {    // Report if OK
			RILlogMessage(NULL, Log_0,
				  "Found %d entries in the OSM file %s",
				  numOsmPoints, osmCnv);
	}
	
	
	status = DALtableGetBuffSize(dalPtrOsm, "OB_TIME", 0, 
								 dataType, 1, numOsmPoints, &bufferSize, status);
	if (status != ISDC_OK){
		RILlogMessage(NULL, Warning_2, 
					  "Couldn't get size of OSM data OBT buffer: %d", status);
		bufferSize = numOsmPoints*sizeof(OBTime);
		status = ISDC_OK;
	}
	
	//	    Make the buffer for the OSM OBT times 
	//
	
	status = DALallocateDataBuffer( (void **)&osmOBTbuffer, bufferSize,
								   status);
	if ( status!= ISDC_OK ){
		RILlogMessage(NULL, Warning_3,
					  "Couldn't allocate data buffer for OSM HK OBTs: %d", status);
	}
	
	
	
	//	    Read all the OBTs for the one revolution
	//
	status = DAL3GENtableGetOBT(dalPtrOsm, "OB_TIME", 1,
								&numRead, osmOBTbuffer, status);
	if ( status!= ISDC_OK ){
		RILlogMessage(NULL, Error_1,
					  "couldn't read in OSM data OBTs: %d", status);
		break;
	}
	
	if ( numRead!=numOsmPoints ){
		RILlogMessage(NULL, Warning_2,
					  "Wrong number OSM data OBT values read in: %ld should be %ld",
					  numRead, numOsmPoints );
		status=2;
	}
		
	
	//
	//	    Make ordering and pruning arrays and initialize them
	//	
	histOrderPtr = NULL;
	histOrderPtr = (int *) malloc( (size_t) numOsmPoints*sizeof(int) );
	hvPruneList = NULL;
	hvPruneList = (int *) malloc( (size_t) numOsmPoints*sizeof(int));
	
	for ( i=0 ; i<numOsmPoints ; i++ ){
		*(histOrderPtr+i) = i;
		*(hvPruneList+i) = good_osm;
	}
	
	//
	//	    Check whether OBTs are disordered or not
	//
	numDisorder = 0;
	for ( i=0 ; i<numOsmPoints-1 ; i++ ){
		status = DAL3GENelapsedOBT( *(osmOBTbuffer+i+1), 
								   *(osmOBTbuffer+i), &elapsedOBT, status);
		if ( status!=ISDC_OK ){
			RILlogMessage(NULL, Warning_2,
						  "%d: Couldn't get elapsed OBT for OSM data entry %ld",
						  status, i);
		}else{
			if ( elapsedOBT<0 ){
				numDisorder++;
			}
		}
	}
	
	 
	//
	//	    If there is some disordering then correct it
	//
	if ( numDisorder > 0 ){
		RILlogMessage(NULL, Log_1,
					  "OSM HK data table disordered in %d places, so sorting it",
					  numDisorder);
		
		status = j_sort(osmOBTbuffer, histOrderPtr, 0, 
							   numOsmPoints-1, status);
		
		/*  Call with values 0 and numOsmPoints-1 refers to beginning
		 and ending row numbers to be sorted.
		 N.B. After call to j_sort the data in osmOBTbuffer is
		 ordered chronologically starting with the least recent time.
		 The histOrderPtr contains the list of array indices by which
		 subsequent OSM table columns should be ordered. i.e.
		 OrderedQuantityArray[n] = DisorderedArray[ *(histOrderPtr+n) ]
		 */
		
		/*	re-seed random generator to get consistent results for sorted
		 and unsorted input OSM data tables
		 */
		DAL3GENrandomSeed(310362);
	}
	
	/*	    Check that the read-in OBTs make sense and are useable
	 NB hvPruneList is using time-ordered OBTs and hence is
	 itself time-ordered
	 */
	nZero=0;
	nNan=0;
	for ( iTest=0 ; iTest<numOsmPoints ; iTest++ ){
		if( isnan( (double) *(osmOBTbuffer+iTest) )!=0){
			hvPruneList[iTest] = obt_nan;
 		    nNan++;
		}else if ( *(osmOBTbuffer+iTest)==DAL3_NO_OBTIME ){
			hvPruneList[iTest] = obt_zero;
			nZero++;
		}
	}
	
	if ( (nNan>0) || (nZero>0) ){
		RILlogMessage(NULL, Warning_3,
					  "%ld OBT values read from OSM data are NaNs", nNan);
		RILlogMessage(NULL, Warning_3,
					  "%ld OSM data OBT values are missing", nZero);
		if ( (nNan+nZero)> numOsmPoints/10 ){
		    RILlogMessage(NULL, Error_1,
						  "Too many unuseable HK OBT values");
			break;
		}
	}
	
		
	// Now the OBT times are read and ordered. 
	//   Make buffer for HV__DELTA column -first find buffer size
	//	
	status = DALtableGetColStruct(dalPtrOsm, 0, &numOsmPoints,
								  (char*) "HV__DELTA", &diskType, &binSize, &varLen, status);
	if (status != ISDC_OK){
		RILlogMessage(NULL, Warning_3,
					  "GetColStruct failed for OSM data HV__DELTA: %d", status);
	}
	
	
	nStart = 1;
	nStop = numOsmPoints;
	status = DALtableGetBuffSize(dalPtrOsm, "HV__DELTA", 0,
								 diskType, nStart, nStop, &bufferSize, status);
	if (status != ISDC_OK){
		RILlogMessage(NULL, Error_1,
					  "DALtableGetBuffSize failed for HV__DELTA: %d", status);
		break;
	}
		
	//	Make HV__DELTA buffer
	//	
	status = DALallocateDataBuffer( (void **)&hvDeltaBuffer, bufferSize,
								   status);
	if ( status!=ISDC_OK ){
		RILlogMessage(NULL, Error_1,
					  "Could not allocate data buffer for HV__DELTA data: %d", 
					  status);
		break;
	}
	

	//	Read in the HV__DELTA data 
	//
	status = DALtableGetCol(dalPtrOsm, "HV__DELTA", 0, 
							&diskType, &numRead, hvDeltaBuffer, status );
	if (status != ISDC_OK){
		RILlogMessage(NULL, Warning_3,
					  "DALtableGetCol failed for OSM data column HV__DELTA: %d.",
					  status);
	}
	if ( numRead!=numOsmPoints ){
		RILlogMessage(NULL, Warning_3,
					  "Wrong no. OSM HV__DELTA values read (%ld vs. %ld)",
					  numRead, numOsmPoints );
	}
	
	if (J_CHATTY_VERBOSE<= chatter) {
		RILlogMessage(NULL, Log_0, "first 10 values read from OSM file are:");
		for ( i=0 ; i<10 ; i++){
			RILlogMessage(NULL, Log_0, "HV__DELTA %8.4g",
				   hvDeltaBuffer[i]);
		}
	}
	
    
	/*	Check for NANs and invalid entries in the HV__DELTA values 
	 NB: hvPruneList is OBT-ordered but the HV_delta buffer is not.  
	 */
	nZero=0;
	nNan=0;
	for ( iTest=0 ; iTest<numOsmPoints ; iTest++ ){
		iSort = *(histOrderPtr+iTest);
		if( (isnan( *(hvDeltaBuffer+iSort) )!=0) ){
			*(hvPruneList+iTest) = osm_nan;
			nNan++;
		}
		if ( (*(hvDeltaBuffer+iSort)< osm_hvLOWER) ){
			*(hvPruneList+iTest) = osm_zero;
			nZero++;
		}
		if ( (*(hvDeltaBuffer+iSort)> osm_hvUPPER) ){
			*(hvPruneList+iTest) = osm_zero;
			nZero++;
		}
	}
	 
	
	if ( nZero>0 ){
		RILlogMessage(NULL, Warning_3,
					  "%ld values read from OSM HV data are invalid", 
					  nZero );
	}
	
	numUsePoints = numOsmPoints - nZero - nNan;
	if ( numUsePoints < numOsmPoints/2 ){
		RILlogMessage(NULL, Warning_0,
					  "Many OSM HV__DELTA are not valid or NANs (%ld)", 
					  nZero+nNan);
	}

	  
	//
	//	    Make array of useable, ordered HV data and OBTs
	//
	HVdata =(double *)malloc( (size_t)numUsePoints*sizeof(double));
	for ( i=0 ; i<numUsePoints ; i++ ){
		*(HVdata+i) = 0.00;
	}
	
	bufferSize = numUsePoints*sizeof(OBTime);
	status = DALallocateDataBuffer( (void **)&orderedOBTs, bufferSize,
								   status);
	if ( status!= ISDC_OK ){
		RILlogMessage(NULL, Warning_3,
					  "Couldn't allocate data buffer for ordered HK AVG OBTs: %d", status);
	}
	
	
	
	/* Sort HV data into the time-ordered array HVdata (and orderedOBTs)
	 NB osmOBTbuffer values and hvPruneList are already sorted 
	 but hvDeltaBuffer is not.
	 */
	iTest=0;
	for ( i=0 ; i<numOsmPoints ; i++){
		iSort = (long) *(histOrderPtr+i);
		if ( hvPruneList[i]>0 ){
		    *(orderedOBTs + iTest) = *(osmOBTbuffer +i);
		    *(HVdata + iTest) = (double) *(hvDeltaBuffer+iSort);
			iTest++;
		}
	}	/* End loop over rows of OSM data */
	
	if ( iTest!=numUsePoints ){
		RILlogMessage(NULL, Log_0,
					  "Final number of useable OSM HV entries: %d", 
					  (int) iTest);
		numUsePoints = iTest;
	}
		 
// Now find the periods where HVdata < HVswitch(=200) (i.e. where the instrument is off)
// Compute the elapsedOBT for these periods
// Exclude all SCWs that are inside *twice* this elapsedOBT 
// (the gain is too variable after the switch-on of the instrument)
	
	
	low_index[0] = 0; // every revolution starts with the instruments off.
	numIntervals = 0;
	
	for ( i=0 ; i<numUsePoints-1 ; i++){
		if ((HVdata[i+1]> HVswitch) && (HVdata[i]<= HVswitch) ){
			high_index[numIntervals]=i;
		}
		if ((HVdata[i+1]<= HVswitch) && (HVdata[i]> HVswitch) ){
			numIntervals++;
			low_index[numIntervals] = i;
		}
	}	/* End loop over rows of OSM data */
	
	/***
	SPR 5105: in some cases the hk tables end just before the perigee entrance.
	    Add a check to find out if we are in this case.
	***/
	
	if (high_index[numIntervals] == -1){ //update the last high_index only in the case that there we are left with an open interval.
		high_index[numIntervals]=numUsePoints-1 ; // every revolution ends with the instruments off.
	}
	 
	if (J_CHATTY_VERBOSE<= chatter) {
		RILlogMessage(NULL, Log_0,
				  "The instruments were switched on/off %d times in rev. %d.", 
				  numIntervals+1, revNUM);
	
		RILlogMessage(NULL, Log_0,
				  "The on/off intervals are");
		for ( i=0 ; i<numIntervals+1 ; i++){
			RILlogMessage(NULL, Log_0, "from entry %d  to %d",
						  low_index[i], high_index[i]);
		}
	}
	
	//compute the elapsed time for each interval
	// and compare the scwOBTstart with the current interval
	skipSCW = false;

	 
	for ( i=0 ; i<numIntervals+1 ; i++){
		status = DAL3GENelapsedOBT( orderedOBTs[high_index[i]],
							   orderedOBTs[low_index[i]], &elapsedOBT, status);
		deltaOBT =elapsedOBT ;
		
		status = DAL3GENelapsedOBT( scwOBTstart,
								   orderedOBTs[low_index[i]], &elapsedOBT, status);
		
		if (J_CHATTY_VERBOSE<= chatter) {
			RILlogMessage(NULL, Log_0, 
						  "instruments were switched-off for %lld [OBT time]; elapsed OBT time from initial switch-off to current SCW: %lld",
					  deltaOBT, elapsedOBT);
		}
		
		if ( (elapsedOBT > 0) && (elapsedOBT < LIN_CONST+deltaOBT )){
			skipSCW = true ;
			break ;
		}
	}
 } while(0);	// end of error-control loop
	
	// Close DALpointer to OSM average data

	if (ISDC_OK==status) {
		DALobjectClose(dalPtrOsm,DAL_SAVE,status);
	}
	
	return status;
}

//   j_sort   (derived from j_cor_gain_sort_history.C v.8.1)
//   Purpose:            To sort an instrument gain history table: Reads 
//                      timeList and orderList arrays and sorts orderList
//                      with quicksort algorithm by OBTimes in timeList
/*   Parameter arguments:     
*
*   Name         Data Type       I/O     Description          Valid Range
*
*   timeList    *OBTime        Input/   Pointer to OBTime    DAL defined OBTime
*				Output  Output array is ordered
*   orderList   *int		Input/  Pointer to orderList  0...numPoints
*				Output  Output array contains original indices
*					of the ordered times ( 0 to numEvts-1)
*   nEvtRowStart int		Input	First row in list to begin sort
*					To sort entire list nEvtRowStart=0
*   nEvtRowEnd	 int		Input	Last row in list to end sort
*					To sort whole list nEvtRowEnd=numEvts-1
*   status       int            Input   Usual input status   0 or negative
*                                                      
*/

int j_sort(OBTime* timeList,
		   int *orderList,
		   int  nEvtRowStart,
		   int  nEvtRowEnd,
		   int status)
{
	int  nHiLoPartition=0;
	
	if (nEvtRowStart < nEvtRowEnd) {
		nHiLoPartition = j_partition(timeList, orderList,
									 nEvtRowStart, nEvtRowEnd, status);
		if ( nHiLoPartition<0 ){
			RILlogMessage(NULL, Warning_2,
						  "%d: Couldn't sort OBTs in gain history table", nHiLoPartition);
			return status;
		}
		
		j_sort(timeList, orderList, nEvtRowStart, nHiLoPartition,
			   status);
		
		j_sort(timeList, orderList, nHiLoPartition + 1, nEvtRowEnd,
			   status);
	}
	return status;
}

//
// j_partition
//
int j_partition(OBTime *timeList,
				int *orderList,
				int  nStart,
				int  nEnd,
				int status)
{
	long loCountUp = nStart - 1;
	long hiCountDown = nEnd + 1;
	long nRandom=0;
	
	OBTime pivot=0;
	OBTime timeswap=0;
	OBTime elapsedOBT=0;
	
	int modeswap;
	
	/* Randomizing the "pivot" element to make the runtime of quicksort \
     independent of the input array */
	
	
	nRandom = DAL3GENrandomInt() % (nEnd - nStart + 1) + nStart;
	pivot = timeList[nRandom];
	status=ISDC_OK;
	
	do {
		do {
			hiCountDown--;
			status = DAL3GENelapsedOBT( timeList[hiCountDown], 
									   pivot, &elapsedOBT, status);
			if ( status!=ISDC_OK ){
				RILlogMessage(NULL, Warning_2,
							  "%d: Couldn't sort OBT times of gain history table", status);
				return status;
			}
		} while ( elapsedOBT>0 );
		do {
			loCountUp++;
			status = DAL3GENelapsedOBT( pivot, 
									   timeList[loCountUp], &elapsedOBT, status);
			if ( status!=ISDC_OK ){
				RILlogMessage(NULL, Warning_2,
							  "%d: Couldn't sort OBT times of gain history table", status);
				return status;
			}
		} while ( elapsedOBT>0 );
		if (loCountUp < hiCountDown) {
			
			timeswap = timeList[loCountUp];
			timeList[loCountUp] = timeList[hiCountDown];
			timeList[hiCountDown] = timeswap;
			
			modeswap = orderList[loCountUp];
			orderList[loCountUp] = orderList[hiCountDown];
			orderList[hiCountDown] = modeswap;
			
			
		} else {
			return hiCountDown;
		}
	} while (1);
	
	return status;
}

